/*
 *
 * Phantom Cipher is Copyright (C) 1996 Kaz Kylheku
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be contacted at:
 *
 * Kaz Kylheku
 * 2869 East 14th Avenue
 * Vancouver, B.C.
 * CANADA
 * V5M 2H8
 * email: kaz@ashi.footprints.net
 *
 * $Id: filecrypt.c,v 1.15 1999/10/17 01:30:53 kaz Exp $
 * $Name: phantom_1_1 $
 */

#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include "phantom.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))

static void parse_args(char **argv);
static void usage(void);
static char *get_password(char *buf, char *buf2, char *prompt, size_t len, int confirm);
static int do_encrypt_decrypt(char *infile, char *outfile, char *phrase, char *salt, int encrypt);
static const char *errmsg(void);

static char pass[514];
static char conf[514];
static char salt[514];

static int encrypting;
static char *infile_name;
static char *outfile_name;
static char *key_phrase;
static char *salt_phrase;

int main(int argc, char **argv)
{
    parse_args(argv);

    if (encrypting && !salt_phrase) {
	get_password(salt_phrase = salt, NULL, "salt: ", sizeof salt, 0);
	if (!*salt) {
	    puts("quit");
	    return EXIT_SUCCESS;
	}
    }

    if (!key_phrase) {
	do {
	    get_password(key_phrase = pass, conf, "password: ",
		    sizeof pass, encrypting);
	    if (!*key_phrase) {
		puts("quit");
		return EXIT_SUCCESS;
	    }
	} while (encrypting && strcmp(key_phrase, conf));
    }

    return do_encrypt_decrypt(infile_name, outfile_name, key_phrase, salt_phrase, encrypting);
}

static void parse_args(char **argv)
{
    if (*++argv) {
	if (argv[0][0] == '-') {
	    switch (argv[0][1]) {
	    case 'e':
		encrypting = 1;
		break;
	    case 'd':
		break;
	    case 'h':
	    default:
		usage();
	    }
	    if (!(infile_name = *++argv)) {
		infile_name = "-";
		outfile_name = "-";
		return;
	    } else if (!(outfile_name = *++argv)) {
		outfile_name = "-";
		return;
	    }

	    if (encrypting) {
		if (!(salt_phrase = *++argv))
		    return;
	    }
	    key_phrase = *++argv;
	} else {
	    usage();
	}
    } else {
	usage();
    }
}

static void usage(void)
{
    char *text =
	"\n"
	"Phantom Cipher file encryption utility version %d.%d\n"
	"Copyright 1996 Kaz Kylheku.\n"
	"\n"
	"to encrypt:\n"
	"\n"
	"phantom -e [ <infile> [ <outfile> [ <salt_phrase> [ <key_phrase> ] ] ] ]\n"
	"\n"
	"to decrypt:\n"
	"\n"
	"phantom -d [ <infile> [ <outfile> [ <key_phrase> ] ] ]\n"
	"\n";


    fprintf(stderr, text, PHANTOM_VERSION_MAJOR, PHANTOM_VERSION_MINOR);
    exit(EXIT_FAILURE);
}

static char *get_password(char *buf, char *buf2, char *prompt, size_t size, int confirm)
{
    int term = open("/dev/tty", O_RDWR);
    FILE *term_s;
    struct termios save, new;
    char *fgetsret, *nl;

    if (term < 0) {
	perror("couldn't open /dev/tty");
	exit(EXIT_FAILURE);
    }

    term_s = fdopen(term, "a+");

    if (!term_s) {
	perror("fdopen failed on /dev/tty");
	exit(EXIT_FAILURE);
    }

    tcgetattr(term, &save);
    new = save;
    new.c_lflag &= ~ECHO;
    new.c_lflag |= ECHONL;
    tcsetattr(term, TCSANOW, &new);
    fprintf(term_s, "%s", prompt);
    fflush(term_s);
    fgetsret = fgets(buf, size, term_s);
    fseek(term_s, 0, SEEK_CUR);
    if ((nl = strchr(buf, '\n')))
	*nl = 0;
    if (!*buf)
	confirm = 0;
    if (confirm) {
	fprintf(term_s, "again: ");
	fflush(term_s);
	fgetsret = fgets(buf2, size, term_s);
	fseek(term_s, 0, SEEK_CUR);
	if ((nl = strchr(buf2, '\n')))
	    *nl = 0;
	if (strcmp(buf, buf2))
	    fprintf(term_s, "they don't match\n");
    }
    tcsetattr(term, TCSANOW, &save);
    fclose(term_s);
    return fgetsret;
}

static int do_encrypt_decrypt(char *infile, char *outfile, char *phrase, char *salt, int encrypt)
{
    int exit_status = EXIT_FAILURE;
    pha_key key;
    pha_schedule sched;
    unsigned char buf[1024];
    pha_cblock ivec;

    FILE *in, *out;

    pha_phrase_to_key(phrase, strlen(phrase), &key);

    if (encrypt) 
	pha_phrase_to_cblock(salt, strlen(salt), &ivec);

    pha_schedule_key(&key, &sched);

    errno = 0;

    if (strcmp(infile, "-") == 0) {
	in = stdin;
	infile = "[stdin]";
    } else {
	in = fopen(infile, "rb");
    }

    if (!in) {
	fprintf(stderr, "could not open \"%s\" for reading: %s\n", infile, errmsg());
	goto bail;
    }

    if (strcmp(outfile, "-") == 0) {
	out = stdout;
	outfile = "[stdout]";
    } else {
	out = fopen(outfile, "wb");
    }

    if (!out) {
	fprintf(stderr, "could not open \"%s\" for writing: %s\n", outfile, errmsg());
	goto cleanup_in;
    }

    if (encrypt)  {
	size_t n_in, n_out;

	/* store initial vector	*/

	if (fwrite(&ivec, 1, sizeof ivec, out) != sizeof ivec)
	    goto write_fail;

	while ((n_in = n_out = fread(buf, 1, sizeof buf, in))) {
	    while (n_out % 16)
		buf[n_out++] = n_in % 256;
	    pha_cbc_encipher((pha_cblock *) buf, (pha_cblock *) buf, &ivec,
		&sched, n_out / 16);
	    if (fwrite(buf, 1, n_out, out) != n_out)
		goto write_fail;
	    if (n_in < sizeof buf)
		break;
	}

	if (ferror(in))
	    goto read_fail;

	if (n_in % 16 == 0) {
	    memset(buf, n_in % 256, 16);
	    pha_cbc_encipher((pha_cblock *) buf, (pha_cblock *) buf, &ivec,
		&sched, 1);
	    if (fwrite(buf, 1, 16, out) != 16)
		goto write_fail;
	}
    } else { /* decrypt */
	size_t nbytes, nout;
	pha_cblock save;
	int saved = 0;

	/* get initial vector	*/

	if (fread(&ivec, 1, sizeof ivec, in) != sizeof ivec) {
	    if (ferror(in))
		goto read_fail;
	    else
		goto read_short;
	    goto cleanup;
	}

	while ((nbytes = fread(buf, 1, sizeof buf, in))) {
	    if (saved && fwrite(&save, 1, 16, out) != 16)
		goto write_fail;
	    if (nbytes % 16 != 0) {
		fprintf(stderr, "warning: length of input from \"%s\" is not multiple of cipher block size\n", infile);
		nbytes -= nbytes % 16;
		if (nbytes == 0)
		    goto read_short;
	    }
	    pha_cbc_decipher((pha_cblock *) buf, (pha_cblock *) buf, &ivec,
		&sched, nbytes / 16);
	    if (nbytes < sizeof buf) {
		nout = nbytes - 16 + buf[nbytes-1] % 16;
		if (fwrite(buf, 1, nout, out) != nout)
		    goto write_fail;
		break;
	    } else {
		if (fwrite(buf, 1, nbytes - 16, out) != nbytes - 16)
		    goto write_fail;
		memcpy(&save, buf + nbytes - 16, 16);
		saved = 1;
	    }
	}

	if (ferror(in))
	    goto read_fail;

	if (nbytes == 0) {
	    if (saved) {
		nout = buf[sizeof buf - 1] % 16;
		if (fwrite(&save, 1, nout, out) != nout)
		    goto write_fail;
	    } else {
		goto read_short;
	    }
	}
    }

    exit_status = EXIT_SUCCESS;

    if (0) {
read_fail:
	fprintf(stderr, "read from \"%s\" failed: %s\n", infile, errmsg());
    }

    if (0) {
write_fail:
	fprintf(stderr, "write to \"%s\" failed: %s\n", outfile, errmsg());
    }

    if (0) {
read_short:
	fprintf(stderr, "premature end of input encountered in \"%s\"\n", infile);
    }

cleanup:
    if (in != stdout)
	fclose(out);

cleanup_in:
    if (in != stdin)
	fclose(in);

bail:

    if (exit_status != EXIT_SUCCESS)
	fprintf(stderr, "unsuccessful %s from \"%s\" to \"%s\"\n", (encrypt) ? "encryption" : "decryption", infile, outfile);
    return exit_status;
}

static const char *errmsg(void)
{
    if (errno == 0)
	return "reason unknown";
    return strerror(errno);
}

